@@ -39,10 +39,10 @@
 #define MAX_EXTRA_CACHED_CERTS 1000
 
 
-/* A certificate cache item.  This consists of a the ksba cert object
+/* A certificate cache item.  This consists of a the KSBA cert object
    and some meta data for easier lookup.  We use a hash table to keep
-   track of all items and use sue the randomly distributed first byte
-   of the fingerprint directly as the hash which makes it pretty easy. */
+   track of all items and use the (randomly distributed) first byte of
+   the fingerprint directly as the hash which makes it pretty easy. */
 struct cert_item_s
 {
   struct cert_item_s *next; /* Next item with the same hash value. */
@@ -117,7 +117,7 @@ compare_serialno (ksba_sexp_t serial1, ksba_sexp_t serial2 )
   unsigned long n_a, n_b;
 
   if (!a || !b)
-    return (!a && !b);
+    return !(!a && !b);
 
   n_a = strtoul (a+1, (char**)&endp, 10);
   a = endp;
@@ -126,10 +126,10 @@ compare_serialno (ksba_sexp_t serial1, ksba_sexp_t serial2 )
   if (*a != ':' || *b != ':')
     {
       log_debug ("invalid S-exp in compare_serialno\n");
-      return 0;
+      return 1;
     }
   if (n_a != n_b)    
-    return 0;
+    return 1;
   a++;
   b++;
 
@@ -138,6 +138,14 @@ compare_serialno (ksba_sexp_t serial1, ksba_sexp_t serial2 )
   return !!n_a;
 }
 
+/* An alias for the above function to be used with S-expressions
+   similar to serial numbers. */
+int
+cmp_simple_canon_sexp (ksba_sexp_t a, ksba_sexp_t b)
+{
+  return compare_serialno (a, b);
+}
+
 
 /* Compute the fingerprint of the certificate CERT and put it into
    the 20 bytes large buffer DIGEST.  Return address of this buffer.  */
@@ -353,22 +361,27 @@ load_certs_from_dir (const char *dirname, int are_trusted)
         }
 
       err = put_cert (cert, 1, are_trusted);
-      ksba_cert_release (cert);
       if (gpg_err_code (err) == GPG_ERR_DUP_VALUE)
         log_info (_("certificate `%s' already cached\n"), fname);
       else if (!err)
         {
-          log_info (_("certificate `%s' loaded\n"), fname);
-          if (are_trusted || opt.verbose)
+          if (are_trusted)
+            log_info (_("trusted certificate `%s' loaded\n"), fname);
+          else
+            log_info (_("certificate `%s' loaded\n"), fname);
+          if (opt.verbose)
             {
               p = get_fingerprint_hexstring_colon (cert);
-              log_info (_("SHA1 fingerprint = %s\n"), p);
+              log_info (_("  SHA1 fingerprint = %s\n"), p);
               xfree (p);
+
+              cert_log_name (_("  name ="), cert);
             }
         }
       else
         log_error (_("error loading certificate `%s': %s\n"),
                      fname, gpg_strerror (err));
+      ksba_cert_release (cert);
     }
 
   xfree (fname);
@@ -589,7 +602,7 @@ get_cert_byissuer (const char *issuer_dn, unsigned int seq)
 }
 
 
-/* Return the certificate matching SUBJEC_DN.  SEQ should initially be
+/* Return the certificate matching SUBJECT_DN.  SEQ should initially be
    set to 0 and bumped up to get the next subject with that DN. */
 ksba_cert_t
 get_cert_bysubject (const char *subject_dn, unsigned int seq)
@@ -618,18 +631,52 @@ get_cert_bysubject (const char *subject_dn, unsigned int seq)
 
 
 
-
 /* Return the certificate matching ISSUER_DN and SERIALNO; if it is
    not already in the cache, try to find it from other resources.  */
 ksba_cert_t
 find_cert_bysn (ctrl_t ctrl, const char *issuer_dn, ksba_sexp_t serialno)
 {
   gpg_error_t err;
-  ksba_cert_t cert = NULL;
+  ksba_cert_t cert;
   cert_fetch_context_t context = NULL;
+  char *hexsn, *buf;
 
+  /* First check whether it has already been cached.  */
   cert = get_cert_bysn (issuer_dn, serialno);
+  if (cert)
+    return cert;
 
+  /* Ask back to the service requester to return the certificate.
+     This is because we can assume that he already used the
+     certificate while checking for the CRL. */
+  hexsn = serial_hex (serialno);
+  if (!hexsn)
+    {
+      log_error ("serial_hex() failed\n");
+      return NULL;
+    }
+  buf = xtrymalloc (1 + strlen (hexsn) + 1 + strlen (issuer_dn) + 1);
+  if (!buf)
+    {
+      log_error ("can't allocate enough memory: %s\n", strerror (errno));
+      xfree (hexsn);
+      return NULL;
+    }
+  strcpy (stpcpy (stpcpy (stpcpy (buf, "#"), hexsn),"/"), issuer_dn);
+  xfree (hexsn);
+  cert = get_cert_local (ctrl, buf);
+  xfree (buf);
+  if (cert)
+    {
+      cache_cert (cert);
+      return cert; /* Done. */
+    }
+
+  if (DBG_LOOKUP)
+    log_debug ("find_cert_bysn: certificate not returned by caller"
+               " - doing lookup\n");
+
+  /* Retrieve the certificate from external resources. */
   while (!cert)
     {
       ksba_sexp_t sn;
@@ -678,6 +725,7 @@ find_cert_bysn (ctrl_t ctrl, const char *issuer_dn, ksba_sexp_t serialno)
       if (!compare_serialno (serialno, sn))
         {
           ksba_free (sn);
+          ksba_free (issdn);
           cache_cert (cert);
           if (DBG_LOOKUP)
             log_debug ("   found\n");
@@ -685,7 +733,146 @@ find_cert_bysn (ctrl_t ctrl, const char *issuer_dn, ksba_sexp_t serialno)
         }
 
       ksba_free (sn);
+      ksba_free (issdn);
+      ksba_cert_release (cert);
+      cert = NULL;
+    }
 
+  end_cert_fetch (context);
+  return cert;
+}
+
+
+/* Return the certificate matching SUBJECT_DN and (if not NULL)
+   KEYID. If it is not already in the cache, try to find it from other
+   resources.  Note, that the external search does not work for user
+   certificates because the LDAP lookup is on the caCertificate
+   attribute. For our purposes this is just fine.  */
+ksba_cert_t
+find_cert_bysubject (ctrl_t ctrl, const char *subject_dn, ksba_sexp_t keyid)
+{
+  gpg_error_t err;
+  int seq;
+  ksba_cert_t cert = NULL;
+  cert_fetch_context_t context = NULL;
+  ksba_sexp_t subj;
+
+  /* First we check whether the certificate is cached.  */
+  for (seq=0; (cert = get_cert_bysubject (subject_dn, seq)); seq++)
+    {
+      if (!keyid)
+        break; /* No keyid requested, so return the first one found. */
+      if (!ksba_cert_get_subj_key_id (cert, NULL, &subj)
+          && !cmp_simple_canon_sexp (keyid, subj))
+        {
+          xfree (subj);
+          break; /* Found matching cert. */
+        }
+      xfree (subj);
+      ksba_cert_release (cert);
+    }
+  if (cert)
+    return cert; /* Done.  */
+
+
+  /* Ask back to the service requester to return the certificate.
+     This is because we can assume that he already used the
+     certificate while checking for the CRL. */
+  if (keyid)
+    cert = get_cert_local_ski (ctrl, subject_dn, keyid);
+  else
+    {
+      /* In contrast to get_cert_local_ski, get_cert_local uses any
+         passed pattern, so we need to make sure that an exact subject
+         search is done. */
+      char *buf;
+
+      buf = xtrymalloc (1 + strlen (subject_dn) + 1);
+      if (!buf)
+        {
+          log_error ("can't allocate enough memory: %s\n", strerror (errno));
+          return NULL;
+        }
+      strcpy (stpcpy (buf, "/"), subject_dn);
+      cert = get_cert_local (ctrl, buf);
+      xfree (buf);
+    }
+  if (cert)
+    {
+      cache_cert (cert);
+      return cert; /* Done. */
+    }
+
+  if (DBG_LOOKUP)
+    log_debug ("find_cert_bysubject: certificate not returned by caller"
+               " - doing lookup\n");
+
+  /* Locate the certificate using external resources. */
+  while (!cert)
+    {
+      char *subjdn;
+
+      if (!context)
+        {
+          err = ca_cert_fetch (ctrl, &context, subject_dn);
+          if (err)
+            {
+              log_error (_("error fetching certificate by subject: %s\n"),
+                         gpg_strerror (err));
+              break;
+            }
+        }
+      
+      err = fetch_next_ksba_cert (context, &cert);
+      if (err)
+        {
+          log_error (_("error fetching certificate by subject: %s\n"),
+                     gpg_strerror (err) );
+          break;
+        }
+      
+      subjdn = ksba_cert_get_subject (cert, 0);
+      if (strcmp (subject_dn, subjdn))
+        {
+          log_info ("find_cert_bysubject: subject DN does not match\n");
+          ksba_cert_release (cert);
+          cert = NULL;
+          ksba_free (subjdn);
+          continue; 
+        }
+
+
+      if (DBG_LOOKUP)
+        {
+          log_debug ("   considering certificate (/");
+          dump_string (subjdn);
+          log_printf (")\n");
+        }
+      ksba_free (subjdn);
+
+      /* If no key ID has been provided, we return the first match.  */
+      if (!keyid)
+        {
+          cache_cert (cert);
+          if (DBG_LOOKUP)
+            log_debug ("   found\n");
+          break; /* Ready.  */
+        }
+
+      /* With the key ID given we need to compare it.  */
+      if (!ksba_cert_get_subj_key_id (cert, NULL, &subj))
+        {
+          if (!cmp_simple_canon_sexp (keyid, subj))
+            {
+              ksba_free (subj);
+              cache_cert (cert);
+              if (DBG_LOOKUP)
+                log_debug ("   found\n");
+              break; /* Ready.  */
+            }
+        }
+
+      ksba_free (subj);
       ksba_cert_release (cert);
       cert = NULL;
     }
@@ -735,6 +922,7 @@ find_issuing_cert (ctrl_t ctrl, ksba_cert_t cert, ksba_cert_t *r_cert)
   ksba_cert_t issuer_cert = NULL;
   ksba_name_t authid;
   ksba_sexp_t authidno;
+  ksba_sexp_t keyid;
 
   *r_cert = NULL;
 
@@ -748,7 +936,7 @@ find_issuing_cert (ctrl_t ctrl, ksba_cert_t cert, ksba_cert_t *r_cert)
 
   /* First we need to check whether we can return that certificate
      using the authorithyKeyIdentifier.  */
-  err = ksba_cert_get_auth_key_id (cert, NULL, &authid, &authidno);
+  err = ksba_cert_get_auth_key_id (cert, &keyid, &authid, &authidno);
   if (err)
     {
       log_info (_("error getting authorityKeyIdentifier: %s\n"),
@@ -761,27 +949,46 @@ find_issuing_cert (ctrl_t ctrl, ksba_cert_t cert, ksba_cert_t *r_cert)
         {
           issuer_cert = find_cert_bysn (ctrl, s, authidno);
         }
+      if (!issuer_cert && keyid)
+        {
+          /* Not found by issuer+s/n.  Now that we have an AKI
+             keyIdentifier look for a certificate with a matching
+             SKI. */
+          issuer_cert = find_cert_bysubject (ctrl, issuer_dn, keyid);
+        }
       /* Print a note so that the user does not feel too helpless when
          an issuer certificate was found and gpgsm prints BAD
          signature because it is not the correct one. */
       if (!issuer_cert)
         {
-          log_info ("issuer certificate (#");
-          dump_serial (authidno);
-          log_printf ("/");
-          dump_string (s);
-          log_printf (") not found\n");
+          log_info ("issuer certificate ");
+          if (keyid)
+            {
+              log_printf ("{");
+              dump_serial (keyid);
+              log_printf ("} ");
+            }
+          if (authidno)
+            {
+              log_printf ("(#");
+              dump_serial (authidno);
+              log_printf ("/");
+              dump_string (s);
+              log_printf (") ");
+            }
+          log_printf ("not found using authorityKeyIdentifier\n");
         }
       ksba_name_release (authid);
       xfree (authidno);
+      xfree (keyid);
     }
 
-  /* If this did not work, try just with the issuer's name.  */
+  /* If this did not work, try just with the issuer's name and assume
+     that there is only one such certificate.  We only look into our
+     cache then. */
   if (err || !issuer_cert)
     {
-      /* FIXME: we need to save a sequence number here for later
-         tries.  */
-      issuer_cert = get_cert_byissuer (issuer_dn, 0);
+      issuer_cert = get_cert_bysubject (issuer_dn, 0);
     }
 
  leave:  
