@@ -124,7 +124,7 @@ static ARGPARSE_OPTS opts[] = {
   { oDebugLevel, "debug-level",2,
                                N_("|LEVEL|set the debugging level to LEVEL")},
   { oNoDetach, "no-detach" ,0, N_("do not detach from the console")},
-  { oLogFile,  "log-file"  ,2, N_("|FILE|write logs to FILE")},
+  { oLogFile,  "log-file"  ,2, N_("|FILE|write server mode logs to FILE")},
   { oBatch   , "batch"     ,0, N_("run without asking a user")},
   { oForce,    "force"     ,0, N_("force loading of outdated CRLs")},
   { oAllowOCSP, "allow-ocsp",0,N_("allow sending OCSP requests")},
@@ -171,6 +171,11 @@ static ARGPARSE_OPTS opts[] = {
   { oLDAPWrapperProgram, "ldap-wrapper-program", 2, "@"},
   { oHTTPWrapperProgram, "http-wrapper-program", 2, "@"},
   { oHonorHTTPProxy,     "honor-http-proxy", 0, "@" },
+
+  { 302, NULL, 0, N_(
+  "@\n(See the \"info\" manual for a complete listing of all commands and options)\n"
+                    )},
+
   {0}
 };
 
