@@ -132,7 +132,7 @@ http_wait_response (http_t hd, unsigned int *ret_status)
 
   http_start_data (hd);  /* Make sure that we are in the data. */
 
-  /* We dup the socket, to cope with thenfact that fclose closes the
+  /* We dup the socket, to cope with the fact that fclose closes the
      underlying socket. */
   hd->sock = dup (hd->sock); 
   if (hd->sock == -1)
@@ -428,6 +428,7 @@ parse_tuple (unsigned char *string)
     {
       tuple->value = p + strlen (p);
       tuple->valuelen = 0;
+      tuple->no_value = 1; /* Explicitly mark that we have seen no '='. */
     }
   else /* name and value */
     {
@@ -547,8 +548,11 @@ build_rel_path (parsed_uri_t uri)
     {
       n++;                      /* '?'/'&' */
       n += insert_escapes (NULL, r->name, "%;?&=");
-      n++;                      /* '=' */
-      n += insert_escapes (NULL, r->value, "%;?&=");
+      if (!r->no_value)
+        {
+          n++;                      /* '=' */
+          n += insert_escapes (NULL, r->value, "%;?&=");
+        }
     }
   n++;
 
@@ -564,10 +568,13 @@ build_rel_path (parsed_uri_t uri)
       *p++ = r == uri->query ? '?' : '&';
       n = insert_escapes (p, r->name, "%;?&=");
       p += n;
-      *p++ = '=';
-      /* TODO: Use valuelen. */
-      n = insert_escapes (p, r->value, "%;?&=");
-      p += n;
+      if (!r->no_value)
+        {
+          *p++ = '=';
+          /* TODO: Use valuelen. */
+          n = insert_escapes (p, r->value, "%;?&=");
+          p += n;
+        }
     }
   *p = 0;
   return rel_path;
@@ -666,7 +673,7 @@ static gpg_error_t
 parse_response (http_t hd)
 {
   unsigned char *line, *p, *p2;
-  unsigned maxlen, len;
+  size_t maxlen, len;
 
   /* Wait for the status line. */
   do
@@ -974,22 +981,30 @@ main (int argc, char **argv)
   printf ("Path  : %s\n", uri->path);
   for (r = uri->params; r; r = r->next)
     {
-      printf ("Params: %s=%s", r->name, r->value);
-      if (strlen (r->value) != r->valuelen)
-        printf (" [real length=%d]", (int) r->valuelen);
+      printf ("Params: %s", r->name);
+      if (!r->no_value)
+        {
+          printf ("=%s", r->value);
+          if (strlen (r->value) != r->valuelen)
+            printf (" [real length=%d]", (int) r->valuelen);
+        }
       putchar ('\n');
     }
   for (r = uri->query; r; r = r->next)
     {
-      printf ("Query : %s=%s", r->name, r->value);
-      if (strlen (r->value) != r->valuelen)
-        printf (" [real length=%d]", (int) r->valuelen);
+      printf ("Query : %s", r->name);
+      if (!r->no_value)
+        {
+          printf ("=%s", r->value);
+          if (strlen (r->value) != r->valuelen)
+            printf (" [real length=%d]", (int) r->valuelen);
+        }
       putchar ('\n');
     }
   http_release_parsed_uri (uri);
   uri = NULL;
 
-  rc = http_open_document (&hd, *argv, 0);
+  rc = http_open_document (&hd, *argv, HTTP_FLAG_NO_SHUTDOWN, NULL);
   if (rc)
     {
       log_error ("can't get `%s': %s\n", *argv, gpg_strerror (rc));
