@@ -26,6 +26,7 @@ struct uri_tuple_s {
   char  *value;         /* A pointer to value (a Nul is always appended). */
   size_t valuelen;	/* The real length of the value; we need it
 			   because the value may contain embedded Nuls */
+  int no_value;         /* True if no value has been given inthe URL */
 };
 typedef struct uri_tuple_s *uri_tuple_t;
 
@@ -66,7 +67,7 @@ struct http_context_s {
   parsed_uri_t uri;
   http_req_t req_type;
   unsigned char *buffer;     /* Line buffer. */
-  unsigned buffer_size;
+  size_t buffer_size;
   unsigned int flags;
 };
 typedef struct http_context_s *http_t;
