@@ -27,6 +27,8 @@
 #include "dirmngr.h"
 #include "misc.h"
 #include "http.h"
+#include "validate.h"
+#include "certcache.h"
 
 #include "ocsp.h"
 
@@ -112,8 +114,13 @@ do_ocsp_request (ctrl_t ctrl, ksba_ocsp_t ocsp, gcry_md_hd_t md,
   struct http_context_s http;
   ksba_ocsp_response_status_t response_status;
   const char *t;
-  
-    
+
+  if (opt.disable_http)
+    {
+      log_error (_("OCSP request not possible due to disabled HTTP\n"));
+      return gpg_error (GPG_ERR_NOT_SUPPORTED);
+    }
+
   err = ksba_ocsp_add_target (ocsp, cert, issuer_cert);
   if (err)
     {
@@ -139,7 +146,9 @@ do_ocsp_request (ctrl_t ctrl, ksba_ocsp_t ocsp, gcry_md_hd_t md,
       return err;
     }
 
-  err = http_open (&http, HTTP_REQ_POST, url, 0);
+  err = http_open (&http, HTTP_REQ_POST, url,
+                   opt.honor_http_proxy? HTTP_FLAG_TRY_PROXY:0,
+                   opt.http_proxy);
   if (err)
     {
       log_error (_("error connecting to `%s': %s\n"), url, gpg_strerror (err));
@@ -236,29 +245,38 @@ do_ocsp_request (ctrl_t ctrl, ksba_ocsp_t ocsp, gcry_md_hd_t md,
 static gpg_error_t
 validate_responder_cert (ctrl_t ctrl, ksba_cert_t cert)
 {
+  gpg_error_t err;
   char *fpr;
 
-  /* We avoid duplicating the entire certificate validation code from
-     gpgsm here.  Because we have no way calling back to the client
-     and letting it compute the validity, we use the ugly hack of
-     telling the client that the response will only be valid if the
-     certificate given in this status message is valid. 
+  if (opt.system_daemon)
+    {
+      err = validate_cert_chain (ctrl, cert, NULL, VALIDATE_MODE_OCSP);
+    }
+  else
+    {
+      /* We avoid duplicating the entire certificate validation code
+         from gpgsm here.  Because we have no way calling back to the
+         client and letting it compute the validity, we use the ugly
+         hack of telling the client that the response will only be
+         valid if the certificate given in this status message is
+         valid.
 
-     Note, that in theory we could simply ask the client via an
-     inquire to validate a certificate but this might involve calling
-     DirMngr again recursivly - we can't do that as of now (neither
-     DirMngr nor gpgsm have the ability for concurrent access to
-     DirMngr.
-   */
+         Note, that in theory we could simply ask the client via an
+         inquire to validate a certificate but this might involve
+         calling DirMngr again recursivly - we can't do that as of now
+         (neither DirMngr nor gpgsm have the ability for concurrent
+         access to DirMngr.   */
 
-  /* FIXME: We should cache this certificate locally, so that the next
-     call to dirmngr won't need to look it up - if this works at
-     all. */
-  fpr = get_fingerprint_hexstring (cert);
-  dirmngr_status (ctrl, "ONLY_VALID_IF_CERT_VALID", fpr, NULL);
-  xfree (fpr);
+      /* FIXME: We should cache this certificate locally, so that the next
+         call to dirmngr won't need to look it up - if this works at
+         all. */
+      fpr = get_fingerprint_hexstring (cert);
+      dirmngr_status (ctrl, "ONLY_VALID_IF_CERT_VALID", fpr, NULL);
+      xfree (fpr);
+      err = 0;
+    }
 
-  return 0;
+  return err;
 }
 
 
@@ -330,7 +348,9 @@ check_signature (ctrl_t ctrl,
     {
       /* We should use the default OCSP reponder's certificate.  Get
          it from the fingerprint. */
-      cert = get_cert_local (ctrl, signer_fpr);
+      cert = get_cert_byhexfpr (signer_fpr);
+      if (!cert)
+        cert = get_cert_local (ctrl, signer_fpr);
       if (cert)
         {
           err = check_signature_core (ctrl, cert, s_sig, s_hash);
@@ -369,17 +389,20 @@ check_signature (ctrl_t ctrl,
 }
 
 
+/* Check whether the certificate either given by fingerprint CERT_FPR
+   or directly through the CERT object is valid by running an OCSP
+   transaction.  */
 gpg_error_t
-ocsp_isvalid (ctrl_t ctrl, const char *cert_fpr)
+ocsp_isvalid (ctrl_t ctrl, ksba_cert_t cert, const char *cert_fpr)
 {
   gpg_error_t err;
   ksba_ocsp_t ocsp = NULL;
-  ksba_cert_t cert = NULL;
   ksba_cert_t issuer_cert = NULL;
   ksba_sexp_t sigval = NULL;
   gcry_sexp_t s_sig = NULL;
   ksba_isotime_t current_time;
   ksba_isotime_t this_update, next_update, revocation_time, produced_at;
+  ksba_isotime_t tmp_time;
   ksba_status_t status;
   ksba_crl_reason_t reason;
   char *url_buffer = NULL;
@@ -390,6 +413,38 @@ ocsp_isvalid (ctrl_t ctrl, const char *cert_fpr)
   ksba_name_t name;
   const char *default_signer = NULL;
 
+  /* Get the certificate.  */
+  if (cert)
+    {
+      ksba_cert_ref (cert);
+
+      err = find_issuing_cert (ctrl, cert, &issuer_cert);
+      if (err)
+        {
+          log_error (_("issuer certificate not found: %s\n"), 
+                     gpg_strerror (err));
+          goto leave;
+        }
+    }
+  else
+    {
+      cert = get_cert_local (ctrl, cert_fpr);
+      if (!cert)
+        {
+          log_error (_("caller did not return the target certificate\n"));
+          err = gpg_error (GPG_ERR_GENERAL);
+          goto leave;
+        }
+      issuer_cert = get_issuing_cert_local (ctrl, NULL);
+      if (!issuer_cert)
+        {
+          log_error (_("caller did not return the issuing certificate\n"));
+          err = gpg_error (GPG_ERR_GENERAL);
+          goto leave;
+        }
+    }
+
+  /* Create an OCSP instance.  */
   err = ksba_ocsp_new (&ocsp);
   if (err)
     {
@@ -398,22 +453,7 @@ ocsp_isvalid (ctrl_t ctrl, const char *cert_fpr)
       goto leave;
     }
 
-  /* We don't have any caching yet implemented thus we need to get the
-     certificate right now. */
-  cert = get_cert_local (ctrl, cert_fpr);
-  if (!cert)
-    {
-      log_error (_("caller did not return the target certificate\n"));
-      err = gpg_error (GPG_ERR_GENERAL);
-      goto leave;
-    }
-  issuer_cert = get_issuing_cert_local (ctrl, NULL);
-  if (!issuer_cert)
-    {
-      log_error (_("caller did not return the issuing certificate\n"));
-      err = gpg_error (GPG_ERR_GENERAL);
-      goto leave;
-    }
+
 
   /* Figure out the OCSP responder to use. 
      1. Try to get the reponder from the certificate.
@@ -421,7 +461,7 @@ ocsp_isvalid (ctrl_t ctrl, const char *cert_fpr)
      2. If this fails use the default responder, if any.
    */
   url = NULL;
-  for (idx=0; !url
+  for (idx=0; !url && !opt.ignore_ocsp_service_url
          && !(err=ksba_cert_get_authority_info_access (cert, idx,
                                                        &oid, &name)); idx++)
     {
@@ -508,6 +548,24 @@ ocsp_isvalid (ctrl_t ctrl, const char *cert_fpr)
                  gpg_strerror (err));
       goto leave;
     }
+
+  /* In case the certificate has been revoked, we better invalidate
+     our cached validation status. */
+  if (status == KSBA_STATUS_REVOKED)
+    {
+      time_t validated_at = 0; /* That is: No cached validation available. */
+      err = ksba_cert_set_user_data (cert, "validated_at",
+                                     &validated_at, sizeof (validated_at));
+      if (err)
+        {
+          log_error ("set_user_data(validated_at) failed: %s\n",
+                     gpg_strerror (err)); 
+          err = 0; /* The certificate is anyway revoked, and that is a
+                      more important message than the failure of our
+                      cache. */
+        }
+    }
+
   
   if (opt.verbose)
     {
@@ -552,16 +610,19 @@ ocsp_isvalid (ctrl_t ctrl, const char *cert_fpr)
   if (*next_update && strcmp (next_update, current_time) < 0 )
     {
       log_error (_("OCSP responder returned an too old status\n"));
+      log_info ("now: %s  next_update: %s\n", current_time, next_update);
       if (!err)
         err = gpg_error (GPG_ERR_TIME_CONFLICT);
     }
   /* Check that THIS_UPDATE is not too far back in the past.  We
      currently use 3 hours (the extra 10 minutes are for the time
      adjust above). */
+  copy_time (tmp_time, this_update);
   add_isotime (this_update, 3 * 60 * 60 + 10 * 60);
   if (!*this_update || strcmp (this_update, current_time) < 0 )
     {
       log_error (_("OCSP responder returned a non-current status\n"));
+      log_info ("now: %s  this_update: %s\n", current_time, tmp_time);
       if (!err)
         err = gpg_error (GPG_ERR_TIME_CONFLICT);
     }
