@@ -838,7 +838,7 @@ cmd_listcrls (assuan_context_t ctx, char *line)
 /* CACHECERT 
 
    Put a certificate into the internal cache.  This command might be
-   usedful if a client knows in advance certificates required for a
+   useful if a client knows in advance certificates required for a
    test and wnats to make sure they get added to the internal cache.
    It is also helpful for debugging.  To get the actual certificate,
    this command immediately inquires it using
@@ -891,7 +891,7 @@ cmd_cachecert (assuan_context_t ctx, char *line)
 
 /* VALIDATE 
 
-   Validate a certificate using the certificate validationj fucntion
+   Validate a certificate using the certificate validation fucntion
    used internally by dirmngr.  This command is only useful for
    debugging.  To get the actual certificate, this command immediately
    inquires it using
@@ -910,7 +910,7 @@ cmd_validate (assuan_context_t ctx, char *line)
   assuan_error_t ae;
   unsigned char *value = NULL;
   size_t valuelen; 
-      
+    
   ae = assuan_inquire (ctrl->server_local->assuan_ctx, "TARGETCERT",
                        &value, &valuelen, MAX_CERT_LENGTH);
   if (ae)
@@ -932,7 +932,23 @@ cmd_validate (assuan_context_t ctx, char *line)
   if(err)
     goto leave;
 
-  err = validate_cert_chain (ctrl, cert, NULL, VALIDATE_MODE_CRL_RECURSIVE);
+  /* If we have this certificate already in our cache, use the cached
+     version for validation because this will take care of any cached
+     results. */
+  { 
+    unsigned char fpr[20];
+    ksba_cert_t tmpcert;
+
+    cert_compute_fpr (cert, fpr);
+    tmpcert = get_cert_byfpr (fpr);
+    if (tmpcert)
+      {
+        ksba_cert_release (cert);
+        cert = tmpcert;
+      }
+  }
+
+  err = validate_cert_chain (ctrl, cert, NULL, VALIDATE_MODE_CERT);
 
  leave:
   if (err)
