@@ -1,6 +1,6 @@
 /* validate.c - Validate a certificate chain.
  *	Copyright (C) 2001, 2003, 2004 Free Software Foundation, Inc.
- *      Copyright (C) 2004 g10 Code GmbH
+ *      Copyright (C) 2004, 2006 g10 Code GmbH
  *
  * This file is part of DirMngr.
  *
@@ -35,8 +35,8 @@
 #include "validate.h"
 #include "misc.h"
 
-/* While running the validation function we need to keep tarckl of the
-   cretificates and the validation outcome of each.  We use this type
+/* While running the validation function we need to keep track of the
+   certificates and the validation outcome of each.  We use this type
    for it.  */
 struct chain_item_s
 {
@@ -192,16 +192,7 @@ check_revocations (ctrl_t ctrl, chain_item_t chain)
   chain_item_t ci;
 
   assert (ctrl->check_revocations_nest_level >= 0);
-
   assert (chain);
-  if (!chain->next)
-    {
-      /* We only have one certificate in the chain.  This is the
-         traget and the Root CA certificate.  We can't do a
-         revocation check for the root certificate because this will
-         in most cases lead to a catch-22.  */
-      return 0; /* Target certificate is a trusted root certificate.  */
-    }
 
   if (ctrl->check_revocations_nest_level > 10)
     {
@@ -214,13 +205,27 @@ check_revocations (ctrl_t ctrl, chain_item_t chain)
   for (ci=chain; ci; ci = ci->next)
     {
       assert (ci->cert);
+      if (ci == chain)
+        {
+          /* It does not make sense to check the root certificate for
+             revocations.  In almost all cases this will lead to a
+             catch-22 as the root certificate is the final trust
+             anchor for the certificates and the CRLs.  We expect the
+             user to remove root certificates from the list of trusted
+             certificates in case they have been revoked. */
+          if (opt.verbose)
+            cert_log_name (_("not checking CRL for"), ci->cert);
+          continue; 
+        }
       
-      err = crl_cache_cert_isvalid (ctrl, chain->cert, 0);
+      if (opt.verbose)
+        cert_log_name (_("checking CRL for"), ci->cert);
+      err = crl_cache_cert_isvalid (ctrl, ci->cert, 0);
       if (gpg_err_code (err) == GPG_ERR_NO_CRL_KNOWN)
         {
-          err = crl_cache_reload_crl (ctrl, chain->cert);
+          err = crl_cache_reload_crl (ctrl, ci->cert);
           if (!err)
-            err = crl_cache_cert_isvalid (ctrl, chain->cert, 0);
+            err = crl_cache_cert_isvalid (ctrl, ci->cert, 0);
         }
       switch (gpg_err_code (err))
         {
@@ -273,31 +278,68 @@ validate_cert_chain (ctrl_t ctrl, ksba_cert_t cert, ksba_isotime_t r_exptime,
   *exptime = 0;
 
   if (!opt.system_daemon)
-    return 0; /* For backward compatibility we only do this in daemon
-                 mode.  */
-
-  /* Get the current time. */
-  get_isotime (current_time);
+    {
+      /* For backward compatibility we only do this in daemon mode.  */
+      log_info (_("running in compatibility mode - "
+                  "certificate chain not checked!\n"));
+      return 0; /* Okay. */
+    }
 
   if (DBG_X509)
     dump_cert ("subject", cert);
   
+  /* May the target certificate be used for this purpose?  */
+  switch (mode)
+    {
+    case VALIDATE_MODE_OCSP:
+      err = cert_use_ocsp_p (cert);
+      break;
+    case VALIDATE_MODE_CRL:
+    case VALIDATE_MODE_CRL_RECURSIVE:
+      err = cert_use_crl_p (cert);
+      break;
+    default:
+      err = 0;
+      break;
+    }
+  if (err)
+    return err;
+
+  /* If we already validated the certificate not too long ago, we can
+     avoid the excessive computations and lookups unless the caller
+     asked for the expiration time.  */
+  if (!r_exptime)
+    {
+      size_t buflen;
+      time_t validated_at;
+      
+      err = ksba_cert_get_user_data (cert, "validated_at", 
+                                     &validated_at, sizeof (validated_at),
+                                     &buflen);
+      if (err || buflen != sizeof (validated_at) || !validated_at)
+        err = 0; /* Not available or other error. */
+      else
+        {
+          /* If the validation is not older than 30 minutes we are ready. */
+          if (validated_at < get_time () + (30*60))
+            {
+              if (opt.verbose)
+                log_info ("certificate is good (cached)\n");
+              /* Note, that we can't jump to leave here as this would
+                 falsely updated the validation timestamp.  */
+              return 0;
+            }
+        }
+    }
+
+  /* Get the current time. */
+  get_isotime (current_time);
+
+  /* We walk up the chain until we find a trust anchor. */
   subject_cert = cert;
-  maxdepth = 50;
+  maxdepth = 10;
   chain = NULL;
   depth = 0;
-
-  /* May the target certificate be used for OCSP response signing?  */
-  if (mode == VALIDATE_MODE_OCSP)
-    {
-      err = cert_use_ocsp_p (cert);
-    }
-  /* May the certificate be used for CR signing.  */
-  /* FIXME. */
-  if (err)
-    goto leave;
-
-  /* We walk up the the path until we find a trust anchor. */
   for (;;)
     {
       /* Get the subject and issuer name from the current
@@ -360,7 +402,7 @@ validate_cert_chain (ctrl_t ctrl, ksba_cert_t cert, ksba_isotime_t r_exptime,
           }            
       }
 
-      /* Do we have any critical extesnions in the certificate we
+      /* Do we have any critical extensions in the certificate we
          can't handle? */
       err = unknown_criticals (subject_cert);
       if (err)
@@ -430,6 +472,10 @@ validate_cert_chain (ctrl_t ctrl, ksba_cert_t cert, ksba_isotime_t r_exptime,
             ci->next = chain;
             chain = ci;
           }
+
+          if (opt.verbose)
+            log_info ("root certificate is good\n");
+
           break;  /* Okay: a self-signed certicate is an end-point. */
         }
 
@@ -471,7 +517,7 @@ validate_cert_chain (ctrl_t ctrl, ksba_cert_t cert, ksba_isotime_t r_exptime,
         }
 
       /* Now check the signature of the certificate.  Well, we
-         shouldnot delay this until later so that faked certificates
+         should delay this until later so that faked certificates
          can't be turned into a DoS easily.  */
       err = check_cert_sig (issuer_cert, subject_cert);
       if (err)
@@ -555,7 +601,7 @@ validate_cert_chain (ctrl_t ctrl, ksba_cert_t cert, ksba_isotime_t r_exptime,
       }
 
       if (opt.verbose)
-        log_info ("certificate is good\n");
+        log_info (_("certificate is good\n"));
 
       /* Now to the next level up.  */
       subject_cert = issuer_cert;
@@ -571,7 +617,16 @@ validate_cert_chain (ctrl_t ctrl, ksba_cert_t cert, ksba_isotime_t r_exptime,
         err = gpg_error (GPG_ERR_NO_POLICY_MATCH);
     }
 
-  if (!err)
+  if (!err && opt.verbose)
+    {
+      chain_item_t citem;
+
+      log_info (_("certificate chain is good\n"));
+      for (citem = chain; citem; citem = citem->next)
+        cert_log_name ("  certificate", citem->cert);
+    }
+  
+  if (!err && mode != VALIDATE_MODE_CRL)
     { /* Now that everything is fine, walk the chain and check each
          certificate for revocations. 
 
@@ -585,12 +640,37 @@ validate_cert_chain (ctrl_t ctrl, ksba_cert_t cert, ksba_isotime_t r_exptime,
          might become a recursive process and we should better cache
          our validity results to avoid double work.  Far worse a
          catch-22 may happen for an improper setup hierachy and we
-         need a way to break up this deadlock. */
+         need a way to break up such a deadlock. */
       err = check_revocations (ctrl, chain);
     }
 
+  if (!err && opt.verbose)
+    log_info ("target certificate is valid\n");
+  else if (err && opt.verbose)
+    log_info ("target certificate is NOT valid\n");
+
   
  leave:
+  if (!err)
+    {
+      /* With no error we can update the validation cache.  We do this
+         for all certificates in the chain. */
+      chain_item_t citem;
+      time_t validated_at = get_time ();
+
+      for (citem = chain; citem; citem = citem->next)
+        {
+          err = ksba_cert_set_user_data (citem->cert, "validated_at",
+                                         &validated_at, sizeof (validated_at));
+          if (err)
+            {
+              log_error ("set_user_data(validated_at) failed: %s\n",
+                         gpg_strerror (err)); 
+              err = 0;
+            }
+        }
+    }
+
   if (r_exptime)
     copy_time (r_exptime, exptime);
   xfree (issuer);
@@ -789,7 +869,7 @@ check_cert_sig (ksba_cert_t issuer_cert, ksba_cert_t cert)
    for signing, a MODE of 1 checks for encryption, a MODE of 2 checks
    for verification and a MODE of 3 for decryption (just for
    debugging).  MODE 4 is for certificate signing, MODE 5 for OCSP
-   response signing. */
+   response signing, MODE 6 is for CRL signing. */
 static int
 cert_usage_p (ksba_cert_t cert, int mode)
 {
@@ -899,6 +979,15 @@ cert_usage_p (ksba_cert_t cert, int mode)
       return gpg_error (GPG_ERR_WRONG_KEY_USAGE);
     }
 
+  if (mode == 6)
+    {
+      if ((use & (KSBA_KEYUSAGE_CRL_SIGN)))
+        return 0;
+      log_info (_("certificate should have not "
+                  "been used for CRL signing\n"));
+      return gpg_error (GPG_ERR_WRONG_KEY_USAGE);
+    }
+
   if ((use & ((mode&1)?
               (KSBA_KEYUSAGE_KEY_ENCIPHERMENT|KSBA_KEYUSAGE_DATA_ENCIPHERMENT):
               (KSBA_KEYUSAGE_DIGITAL_SIGNATURE|KSBA_KEYUSAGE_NON_REPUDIATION)))
@@ -928,3 +1017,10 @@ cert_use_ocsp_p (ksba_cert_t cert)
   return cert_usage_p (cert, 5);
 }
 
+/* Return 0 if the certificate CERT is usable for signing CRLs. */
+gpg_error_t
+cert_use_crl_p (ksba_cert_t cert)
+{
+  return cert_usage_p (cert, 6);
+}
+
